@@ -1058,23 +1058,15 @@ static void normal_stdin (scriptelem_t const *script, unsigned int scriptlen, si
 
 static void last_stdin (scriptelem_t const *script, unsigned int scriptlen, size_t linelimit, unsigned int gflags)
 {
-  int cont = 1 ;
-  while (cont)
+  for (;;)
   {
     char c ;
     switch (sanitize_read(fd_read(0, &c, 1)))
     {
-      case 0 :
-        cont = 0 ;
-        break ;
+      case 0 : return ;
       case -1 :
         if ((errno != EPIPE) && verbosity) strerr_warnwu1sys("read from stdin") ;
-        if (!indata.len)
-        {
-          prepare_to_exit() ;
-          cont = 0 ;
-          break ;
-        }
+        if (!indata.len) goto eof ;
  addfinalnewline:
         c = '\n' ;
       case 1 :
@@ -1082,8 +1074,7 @@ static void last_stdin (scriptelem_t const *script, unsigned int scriptlen, size
         if (c == '\n')
         {
           script_run(script, scriptlen, indata.s, indata.len - 1, gflags) ;
-          prepare_to_exit() ;
-          cont = 0 ;
+          goto eof ;
         }
         else if (linelimit && indata.len > linelimit)
         {
@@ -1093,6 +1084,8 @@ static void last_stdin (scriptelem_t const *script, unsigned int scriptlen, size
         break ;
     }
   }
+ eof:
+  prepare_to_exit() ;
 }
 
 static inputproc_func_ref handle_stdin = &normal_stdin ;
@@ -1122,14 +1115,15 @@ static inline void processor_died (logdir_t *ldp, int wstat)
   }
 }
 
-static inline void handle_signals (void)
+static inline int handle_signals (void)
 {
+  int e = 0 ;
   for (;;)
   {
     switch (selfpipe_read())
     {
       case -1 : strerr_diefu1sys(111, "selfpipe_read") ;
-      case 0 : return ;
+      case 0 : return e ;
       case SIGALRM :
       {
         unsigned int i = 0 ;
@@ -1145,7 +1139,7 @@ static inline void handle_signals (void)
         if (flagprotect) break ;
       case SIGHUP :
         handle_stdin = &last_stdin ;
-        if (!indata.len) prepare_to_exit() ;
+        if (!indata.len) { prepare_to_exit() ; e = 1 ; }
         break ;
       case SIGCHLD :
       {
@@ -1284,8 +1278,7 @@ int main (int argc, char const *const *argv)
       if (r < 0) strerr_diefu1sys(111, "iopause") ;
       else if (!r) continue ;
 
-      if (x[0].revents & IOPAUSE_READ) handle_signals() ;
-      else if (x[0].revents & IOPAUSE_EXCEPT) strerr_dief1sys(111, "trouble with selfpipe") ;
+      if (x[0].revents & (IOPAUSE_READ | IOPAUSE_EXCEPT) && handle_signals()) continue ;
 
       if (xindex1 && x[xindex1].revents)
       {
